# -*- coding: utf-8 -*-
# Morra project: Morphological parser
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Single-pass forward and backward morphological parsers for POS and FEATS
tagging. Also, wrapper for LEMMA generator of Corpuscula project is added.
"""
from collections import OrderedDict
from copy import deepcopy
import pickle
import random
from random import randint, random as rand
import sys

from corpuscula.corpus_utils import _AbstractCorpus
from corpuscula.utils import LOG_FILE, print_progress
from morra.base_parser import _AveragedPerceptron, BaseParser
from morra.features import Features


class MorphParser(BaseParser):
    """Single-pass forward and backward morphological parsers. Based on
    PerceptronTagger from NLTK extended by CorpusDict methods and generalized
    to features other than POS"""

    def __init__(self, features='RU',
                 guess_pos=None, guess_lemma=None, guess_feat=None):
        super().__init__()
        self.features = Features(lang=features) \
            if isinstance(features, str) else features

        self._guess_pos    = guess_pos
        self._guess_lemma  = guess_lemma
        self._guess_feat   = guess_feat

        self._pos_model        = None
        self._pos_rev_model    = None
        self._lemma_model      = None
        self._feats_model      = None
        self._feats_rev_model  = None
        self._feats_models     = {}
        self._feats_rev_models = {}

    def backup(self):
        """Get current state"""
        o = super().backup()
        o.update({'pos_model_weights'       : self._pos_model.weights
                                                  if self._pos_model       else
                                              None,
                  'pos_rev_model_weights'   : self._pos_rev_model.weights
                                                  if self._pos_rev_model   else
                                              None,
                  'self._lemma_model'       : self._lemma_model,
                  'feats_model_weights'     : self._feats_model.weights
                                                  if self._feats_model     else
                                              None,
                  'feats_rev_model_weights' : self._feats_rev_model.weights
                                                  if self._feats_rev_model else
                                              None,
                  'feats_models_weights'    : {
                      x: y.weights for x, y in self._feats_models.items()
                  },
                  'feats_rev_models_weights': {
                      x: y.weights for x, y in self._feats_rev_models.items()
                  }})
        return o

    def restore(self, o):
        """Restore current state from backup object"""
        super().restore(o)
        (pos_model_weights       ,
         pos_rev_model_weights   ,
         self._lemma_model       ,
         feats_model_weights     ,
         feats_rev_model_weights ,
         feats_models_weights    ,
         feats_rev_models_weights) = \
             [o.get(x) for x in ['pos_model_weights'       ,
                                 'pos_rev_model_weights'   ,
                                 'self._lemma_model'       ,
                                 'feats_model_weights'     ,
                                 'feats_rev_model_weights' ,
                                 'feats_models_weights'    ,
                                 'feats_rev_models_weights']]
        if pos_model_weights:
            self._pos_model = _AveragedPerceptron()
            #self._pos_model.default_class = self._cdict.most_common_tag()
            self._pos_model.weights = pos_model_weights
        else:
            self._pos_model = None
        if pos_rev_model_weights:
            self._pos_rev_model = _AveragedPerceptron()
            self._pos_rev_model.weights = pos_rev_model_weights
        else:
            self._pos_rev_model = None
        if feats_model_weights:
            self._feats_model = _AveragedPerceptron()
            self._feats_model.weights = feats_model_weights
        else:
            self._feats_model = None
        if feats_rev_model_weights:
            self._feats_rev_model = _AveragedPerceptron()
            self._feats_rev_model.weights = feats_rev_model_weights
        else:
            self._feats_rev_model = None
        models = self._feats_models = {}
        if feats_models_weights:
            for feat, weights in feats_models_weights.items():
                model = models[feat] = _AveragedPerceptron()
                #model.default_class = '_'
                model.weights = weights
        models = self._feats_rev_models = {}
        if feats_rev_models_weights:
            for feat, weights in feats_rev_models_weights.items():
                model = models[feat] = _AveragedPerceptron()
                model.weights = weights

    def _save_pos_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._pos_model.weights if self._pos_model else
                        None, f, 2)

    def _load_pos_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._pos_model = _AveragedPerceptron()
                self._pos_model.weights = weights
            else:
                self._pos_model = None

    def _save_pos_rev_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._pos_rev_model.weights if self._pos_rev_model else
                        None, f, 2)

    def _load_pos_rev_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._pos_rev_model = _AveragedPerceptron()
                self._pos_rev_model.weights = weights
            else:
                self._pos_rev_model = None

    def _save_lemma_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._lemma_model, f, 2)

    def _load_lemma_model(self, file_path):
        with open(file_path, 'rb') as f:
            self._lemma_model = pickle.load(f)

    def _save_feats_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._feats_model.weights if self._feats_model else
                        None, f, 2)

    def _load_feats_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._feats_model = _AveragedPerceptron()
                self._feats_model.weights = weights
            else:
                self._feats_model = None

    def _save_feats_rev_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._feats_rev_model.weights
                            if self._feats_rev_model else
                        None, f, 2)

    def _load_feats_rev_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._feats_rev_model = _AveragedPerceptron()
                self._feats_rev_model.weights = weights
            else:
                self._feats_rev_model = None

    def _save_feats_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump(
                (feat, self._feats_models[feat].weights) if feat else
                {x: y.weights for x, y in self._feats_models.items()},
                f, 2
            )

    def _load_feats_models(self, file_path):
        with open(file_path, 'rb') as f:
            o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._feats_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._feats_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def _save_feats_rev_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump(
                (feat, self._feats_rev_models[feat].weights) if feat else
                {x: y.weights for x, y in self._feats_rev_models.items()},
                f, 2
            )

    def _load_feats_rev_models(self, file_path):
        with open(file_path, 'rb') as f:
            o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._feats_rev_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._feats_rev_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def predict_pos(self, sentence, rev=False, inplace=True):
        """Tag the *sentence* with the POS tagger.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param rev: if True, use Reversed POS tagger instead of generic
                    straight one
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        cdict = self._cdict
        model = self._pos_rev_model if rev else self._pos_model
        assert model, \
               'ERROR: Use train_pos({}) prior to prepare {}POS tagger' \
                   .format(*(('rev=True', 'Reversed') if rev else ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        tokens = [[x['FORM'], None]
                      for x in sent
                          if x['FORM'] and '-' not in x['ID']]
        context = [t[0] for t in (reversed(tokens) if rev else tokens)]
        prev, prev2 = self.features.START
        i = 0
        max_i = len(context) - 1
        for token in reversed(sent) if rev else sent:
            wform = token['FORM']
            if wform and '-' not in token['ID']:
                i_ = max_i - i if rev else i
                guess, coef = cdict.predict_tag(wform, isfirst=i_ == 0)
                if self._guess_pos:
                    guess, coef = self._guess_pos(guess, coef, i_,
                                                  tokens, cdict)
                if guess is None or coef < 1.:
                    features = self.features.get_pos_features(
                        i, context, prev, prev2
                    )
                    guess = model.predict(
                        features, suggest=guess, suggest_coef=coef
                    )
                prev2 = prev
                token['UPOS'] = tokens[i_][1] = prev = guess
                i += 1
            else:
                token['UPOS'] = None
        return sentence

    def predict_lemma(self, sentence, inplace=True):
        """Generate lemmata for wforms of the *sentence*.

        :param sentence: sentence in Parsed CoNLL-U format; UPOS field must be 
                         already filled
        :type sentence: list(dict)
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, the new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        cdict = self._cdict
        assert self._lemma_model, \
               'ERROR: Use train_lemma() prior to prepare Lemma generator'
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        tokens = [[x['FORM'], x['UPOS'], None]
                      for x in sent
                          if x['FORM'] and x['UPOS'] and '-' not in x['ID']]
        i = 0
        for token in sent:
            wform, pos = token['FORM'], token['UPOS']
            if wform and pos and '-' not in token['ID']:
                guess, coef = cdict.predict_lemma(wform, pos, isfirst=i == 0)
                if self._guess_lemma:
                    guess, coef = self._guess_lemma(guess, coef, i,
                                                    tokens, cdict)
                token['LEMMA'] = tokens[i][2] = guess
                i += 1
            else:
                token['LEMMA'] = None
        return sentence

    def predict_feats(self, sentence, joint=False, rev=False, feat=None,
                      inplace=True):
        """Tag the *sentence* with the FEATS tagger.

        :param sentence: sentence in Parsed CoNLL-U format; UPOS and LEMMA
                         fields must be already filled
        :type sentence: list(dict)
        :param joint: if True, use joint FEATS model; elsewise, use separate
                      models (default)
        :param rev: if True, use Reversed FEATS tagger instead of generic
                    straight one
        :param feat: name of the feat to tag; if None, then all possible feats
                     will be tagged
        :type feat: str
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        return (
            self._predict_feats_joint if joint else
            self._predict_feats_separate
        )(
            sentence, rev=rev, feat=feat, inplace=inplace
        )

    def _predict_feats_separate(self, sentence, rev=False, feat=None,
                                inplace=True):
        cdict = self._cdict
        models = self._feats_rev_models if rev else self._feats_models
        assert models, \
               'ERROR: Use train_feats(joint=False{}) prior to prepare ' \
               '{}FEATS tagger' \
                   .format(*((', rev=True', 'Reversed') if rev else ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        if not feat:
            for token in sent:
                token['FEATS'] = OrderedDict()
            for feat in cdict.get_feats():
                self._predict_feats_separate(sent, rev=rev, feat=feat,
                                             inplace=True)
        else:
            default_val = '_'
            model = models[feat]
            val_cnt = len(cdict.get_feats()[feat]) - 1
            tokens = [[x['FORM'], x['LEMMA'], x['UPOS'], None]
                          for x in sent
                              if x['FORM'] and x['LEMMA'] and x['UPOS']
                             and '-' not in x['ID']]
            context, lemma_context, pos_context = [list(x) for x in zip(
                *[t[:3] for t in (reversed(tokens) if rev else tokens)]
            )] if tokens else [[]] * 3
            prev, prev2 = self.features.START
            i = 0
            max_i = len(tokens) - 1
            for token in reversed(sent) if rev else sent:
                wform, lemma, pos, feats = token['FORM'], token['LEMMA'], \
                                           token['UPOS'], token['FEATS']
                if wform and lemma and pos and '-' not in token['ID']:
                    i_ = max_i - i if rev else i
                    guess, coef = cdict.predict_feat(feat, wform, lemma, pos)
                    if self._guess_feat:
                        guess, coef = self._guess_feat(guess, coef, i_,
                                                       feat, tokens, cdict)
                    if coef is not None and guess is None:
                        guess = default_val
                    if coef != 1.:
                        features = self.features.get_feat_features(
                            i, context, lemma_context, pos_context,
                            False, val_cnt, prev, prev2
                        )
                        guess = model.predict(
                            features, suggest=guess, suggest_coef=coef
                        )
                    prev2 = prev
                    tokens[i_][3] = prev = guess
                    if guess != default_val:
                        feats[feat] = guess
                    else:
                        feats.pop(feat, None)
                    i += 1
                else:
                    feats.clear()
        return sentence

    def _predict_feats_joint(self, sentence, rev=False, feat=None,
                             inplace=True):
        assert not feat, 'ERROR: feat must be None with joint=True'
        cdict = self._cdict
        model = self._feats_rev_model if rev else self._feats_model
        assert model, \
               'ERROR: Use train_feats(joint=True{}) prior to prepare ' \
               '{}FEATS tagger' \
                   .format(*((', rev=True', 'Reversed') if rev else ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        tokens = [(x['FORM'], x['LEMMA'], x['UPOS'])
                      for x in (reversed(sentence) if rev else sentence)
                          if x['FORM'] and x['LEMMA'] and x['UPOS']
                         and '-' not in x['ID']]
        context, lemma_context, pos_context = \
            [list(x) for x in zip(*[t for t in tokens])] if tokens else \
            [[]] * 3
        prev, prev2 = self.features.START
        i = 0
        for token in reversed(sent) if rev else sent:
            wform, lemma, pos, feats = token['FORM'], token['LEMMA'], \
                                       token['UPOS'], token['FEATS']
            feats.clear()
            if wform and lemma and pos and '-' not in token['ID']:
                features = self.features.get_feat_features(
                    i, context, lemma_context, pos_context,
                    True, 0, prev, prev2
                )
                guess = model.predict(features)
                prev2 = prev
                prev = guess
                feats.clear()
                if guess:
                    for feat, val in [t.split('=') for t in guess.split('|')]:
                        feats[feat] = val
                i += 1
        return sentence

    def predict(self, sentence, pos_rev=False, feats_joint=False,
                feats_rev=False, inplace=True):
        """Tag the *sentence* with the all available taggers.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param pos_rev: if True, use Reversed POS tagger instead of generic
                        straight one
        :param feats_joint: if True, use joint FEATS model; elsewise, use
                            separate models (default)
        :param feats_rev: if True, use Reversed FEATS tagger instead of generic
                          straight one
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        return \
            self.predict_feats(
                self.predict_lemma(
                    self.predict_pos(
                        sentence, rev=pos_rev, inplace=inplace
                    ),
                    inplace=inplace
                ),
                joint=feats_joint, rev=feats_rev, inplace=inplace
            )

    def predict_pos_sents(self, sentences=None, rev=False, inplace=True,
                          save_to=None):
        """Apply ``self.predict_pos()`` to each element of *sentences*.

        :param sentences: a name of a file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param rev: if True, use Reversed POS tagger instead of generic
                    straight one
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_pos(s, rev=rev, inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def predict_lemma_sents(self, sentences=None, inplace=True, save_to=None):
        """Apply ``self.predict_lemma()`` to each element of *sentences*.

        :param sentences: a name of a file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, the new list of
                        sentences will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_lemma(s, inplace=inplace) for s in sentences),
            save_to=save_to
        )

    def predict_feats_sents(self, sentences=None, joint=False, rev=False,
                            feat=None, inplace=True, save_to=None):
        """Apply ``self.predict_feats()`` to each element of *sentences*.

        :param sentences: a name of a file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param joint: if True, use joint FEATS model; elsewise, use separate
                      models (default)
        :param rev: if True, use Reversed FEATS tagger instead of generic
                    straight one
        :param feat: name of the feat to tag; if None, then all feats will be
                     tagged
        :type feat: str
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, the new list of
                        sentences will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_feats(s, joint=joint, rev=rev, feat=feat,
                                    inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def predict_sents(self, sentences=None, pos_rev=False, feats_joint=False,
                      feats_rev=False, inplace=True, save_to=None):
        """Apply ``self.predict()`` to each element of *sentences*.

        :param sentences: a name of the file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param pos_rev: if True, use Reversed POS tagger instead of generic
                        straight one
        :param feats_joint: if True, use joint FEATS model; elsewise, use
                            separate models (default)
        :param feats_rev: if True, use Reversed FEATS tagger instead of generic
                          straight one
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict(s, pos_rev=pos_rev, feats_joint=feats_joint,
                              feats_rev=feats_rev, inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def evaluate_pos(self, gold=None, test=None, rev=False, pos=None,
                     unknown_only=False, silent=False):
        """Score the accuracy of the POS tagger against the *gold* standard.
        Remove POS tags from the *gold* standard text, retag it using the
        tagger, then compute the accuracy score. If *test* is not None, compute
        the accuracy of the *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param rev: if True, use Reversed POS tagger instead of generic
                    straight one
        :param pos: name of the tag to evaluate the tagger; if None, then
                    tagger will be evaluated for all tags
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy score of the tagger against the gold
        :rtype: float
        """
        n = c = nt = ct = w1 = w2 = 0
        if gold is None:
            gold = self._test_corpus
        elif (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
          or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = self._get_corpus(gold, silent=True)
        if test:
            test = self._get_corpus(test, silent=True)
        header = '{}POS{}'.format('Reversed ' if rev else '',
                                  '<<{}>>'.format(pos) if pos else '')
        if not silent:
            print('Evaluate ' + header, file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        cdict = self._cdict
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            test_sent = next(test) if test else \
                        self.predict_pos(gold_sent, rev=rev, inplace=False)
            for j, gold_token in enumerate(gold_sent):
                wform = gold_token['FORM']
                if wform and '-' not in gold_token['ID']:
                    gold_pos = gold_token['UPOS']
                    if not (
                        unknown_only and cdict.wform_isknown(wform,
                                                             tag=gold_pos)
                    ):
                        test_pos = test_sent[j]['UPOS']
                        n += 1
                        c += gold_pos == test_pos
                        if pos and (gold_pos == pos or test_pos == pos):
                            nt += 1
                            if gold_pos == test_pos:
                                ct += 1
                            elif test_pos == pos:
                                w1 += 1
                            else:
                                w2 += 1
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                sp = ' ' * (len(header) - 2)
                print(header + ' total: {}'.format(nt if pos else n),
                      file=LOG_FILE)
                print(sp   + ' correct: {}'.format(ct if pos else c),
                      file=LOG_FILE)
                print(sp   + '   wrong: {}{}'
                                 .format(nt - ct if pos else n - c,
                                         ' [{} excess / {} absent]'
                                             .format(w1, w2) if pos else
                                         ''),
                      file=LOG_FILE)
                if pos:
                    print(sp   + 'Accuracy: {}'
                                     .format(ct / nt if nt > 0 else 1.),
                          file=LOG_FILE)
                print('Total accuracy: {}'.format(c / n if n > 0 else 1.),
                      file=LOG_FILE)
        return c / n if n > 0 else 1.

    def evaluate_lemma(self, gold=None, test=None, unknown_only=False,
                       silent=False):
        """Score the accuracy of the Lemma generator against the *gold*
        standard. Remove lemmata from the *gold* standard text, generate new
        lemmata using the tagger, then compute the accuracy score. If *test* is
        not None, compute the accuracy of the *test* corpus with respect to the
        *gold*.

        :param gold: a corpus of tagged sentences to score the generator on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy score of the generator against the gold
        :rtype: float
        """
        n = c = 0
        if gold is None:
            gold = self._test_corpus
        elif (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
          or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = self._get_corpus(gold, silent=True)
        if test:
            test = self._get_corpus(test, silent=True)
        header = 'LEMMA'
        if not silent:
            print('Evaluate ' + header, file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        cdict = self._cdict
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            test_sent = next(test) if test else \
                        self.predict_lemma(gold_sent, inplace=False)
            for j, gold_token in enumerate(gold_sent):
                wform = gold_token['FORM']
                if wform and '-' not in gold_token['ID']:
                    gold_pos = gold_token['UPOS']
                    if not (
                        unknown_only and cdict.wform_isknown(wform,
                                                             tag=gold_pos)
                    ):
                        n += 1
                        c += gold_token['LEMMA'] == test_sent[j]['LEMMA']
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                sp = ' ' * (len(header) - 2)
                print(header + ' total: {}'.format(n), file=LOG_FILE)
                print(sp   + ' correct: {}'.format(c), file=LOG_FILE)
                print(sp   + '   wrong: {}'.format(n - c), file=LOG_FILE)
                print(sp   + 'Accuracy: {}'.format(c / n if n > 0 else 1.),
                      file=LOG_FILE)
        return c / n if n > 0 else 1.

    def evaluate_feats(self, gold=None, test=None, joint=False, rev=False,
                       feat=None, unknown_only=False, silent=False):
        """Score the accuracy of the FEATS tagger against the *gold* standard.
        Remove feats (or only one specified feat) from the *gold* standard
        text, generate new feats using the tagger, then compute the accuracy
        score. If *test* is not None, compute the accuracy of the *test* corpus
        with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param joint: if True, use joint FEATS model; elsewise, use separate
                      models (default)
        :param rev: if True, use Reversed FEATS tagger instead of generic
                    straight one
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy scores of the tagger against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not;
                 2. by tags: sum of correctly detected feats to sum of all
                    feats that are non-empty in either gold or retagged 
                    sentence
        :rtype: tuple(float, float)
        """
        n = c = nt = ct = 0
        if gold is None:
            gold = self._test_corpus
        elif (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
          or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = self._get_corpus(gold, silent=True)
        if test:
            test = self._get_corpus(test, silent=True)
        header = '{}FEAT{}{}'.format('Reversed ' if rev else '',
                                     '' if feat else 'S',
                                     '-j' if joint else '',
                                     '<<{}>>'.format(feat) if feat else '')
        if not silent:
            print('Evaluate ' + header, file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        cdict = self._cdict
        feat_vals = cdict.get_feats()
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            #gold_sent = self.predict_pos(gold_sent, rev=rev, inplace=False)
            test_sent = next(test) if test else \
                        self.predict_feats(gold_sent, joint=joint, rev=rev,
                                           feat=None if joint else feat,
                                           inplace=False)
            for j, gold_token in enumerate(gold_sent):
                wform, gold_pos = gold_token['FORM'], gold_token['UPOS']
                if wform and '-' not in gold_token['ID']:
                    if not (
                        unknown_only and cdict.wform_isknown(wform,
                                                             tag=gold_pos)
                    ):
                        nf = cf = 0
                        gold_feats = gold_token['FEATS']
                        test_feats = test_sent[j]['FEATS']
                        for feat_ in [feat] if feat else feat_vals:
                            gold_feat = gold_feats.get(feat_)
                            test_feat = test_feats.get(feat_)
                            if gold_feat or test_feat:
                                nf += 1
                                if gold_feat == test_feat:
                                    cf += 1
                        if nf > 0:
                            n += 1
                            c += nf == cf
                            nt += nf
                            ct += cf
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                sp = ' ' * (len(header) - 2)
                print(header + ' total: {} tokens, {} tags'.format(n, nt),
                      file=LOG_FILE)
                print(sp   + ' correct: {} tokens, {} tags'.format(c, ct),
                      file=LOG_FILE)
                print(sp   + '   wrong: {} tokens, {} tags'
                                 .format(n - c, nt - ct),
                      file=LOG_FILE)
                print(sp   + 'Accuracy: {} / {}'
                                 .format( c / n  if n  > 0 else 1.,
                                         ct / nt if nt > 0 else 1.),
                          file=LOG_FILE)
        return c / n if n > 0 else 1., ct / nt if nt > 0 else 1.

    def evaluate(self, gold=None, test=None, pos_rev=False,
                 feats_joint=False, feats_rev=False, feat=None,
                 unknown_only=False, silent=False):
        """Score a joint accuracy of the all available taggers against the
        *gold* standard. Extract wforms from the *gold* standard text, retag it
        using all the taggers, then compute a joint accuracy score. If *test*
        is not None, compute the accuracy of the *test* corpus with respect to
        the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param pos_rev: if True, use Reversed POS tagger instead of generic
                        straight one
        :param feats_joint: if True, use joint FEATS model; elsewise, use
                            separate models (default)
        :param feats_rev: if True, use Reversed FEATS tagger instead of generic
                          straight one
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: joint accuracy scores of the taggers against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not
                 2. by tags: sum of correctly detected tags to sum of all tags
                    that are non-empty in either gold or retagged sentences
        :rtype: tuple(float, float)
        """
        n = c = nt = ct = 0
        if gold is None:
            gold = self._test_corpus
        elif (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
          or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = self._get_corpus(gold, silent=True)
        if test:
            test = self._get_corpus(test, silent=True)
        feat_vals = self._cdict.get_feats()
        if not silent:
            print('Evaluate', file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        cdict = self._cdict
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            test_sent = next(test) if test else \
                        self.predict(gold_sent, pos_rev=pos_rev,
                                     feats_joint=feats_joint,
                                     feats_rev=feats_rev, inplace=False)
            for j, gold_token in enumerate(gold_sent):
                wform = gold_token['FORM']
                if wform and '-' not in gold_token['ID']:
                    gold_pos = gold_token['UPOS']
                    if not (
                        unknown_only and cdict.wform_isknown(wform,
                                                             tag=gold_pos)
                    ):
                        test_token = test_sent[j]
                        nf = cf = 0
                        # LEMMA
                        nf += 1
                        cf += gold_token['LEMMA'] == test_token['LEMMA']
                        # UPOS
                        nf += 1
                        cf += gold_pos == test_token['UPOS']
                        # FEATS
                        gold_feats = gold_token['FEATS']
                        test_feats = test_token['FEATS']
                        for feat_ in [feat] if feat else feat_vals:
                            gold_feat = gold_feats.get(feat_)
                            test_feat = test_feats.get(feat_)
                            if gold_feat or test_feat:
                                nf += 1
                                if gold_feat == test_feat:
                                    cf += 1
                        if nf > 0:
                            n += 1
                            c += nf == cf
                            nt += nf
                            ct += cf
        res = c / n if n > 0 else 1., ct / nt if nt > 0 else 1.
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                print('   total: {} tokens, {} tags'.format(n, nt),
                      file=LOG_FILE)
                print(' correct: {} tokens, {} tags'.format(c, ct),
                      file=LOG_FILE)
                print('   wrong: {} tokens, {} tags'.format(n - c, nt - ct),
                      file=LOG_FILE)
                print('Accuracy: {} / {}'.format(res[0], res[1]),
                      file=LOG_FILE)
        return res

    def train_pos(self, rev=False, epochs=5, no_train_evals=True, seed=None,
                  dropout=None, context_dropout=None):
        """Train a POS tagger from ``self._train_corpus``.

        :param rev: if True, train Reversed POS tagger instead of generic
                    straight one
        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of POS tags to be randomly replaced
                                after predict to random POS tags to prevent
                                overfitting
        :type context_dropout: float
        """
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)

        model = _AveragedPerceptron(default_class=cdict.most_common_tag())
        if rev:
            self._pos_rev_model = model
        else:
            self._pos_model = model

        header = '{}POS'.format('Reversed ' if rev else '')
        tags = sorted(cdict.get_tags())
        last_tag_idx = len(tags) - 1
        print(tags, file=LOG_FILE)
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['UPOS'])
                              for x in sentence
                                  if x['FORM'] and '-' not in x['ID']]
                tokens_ = [[t[0], None] for t in tokens]
                if rev:
                    tokens = tokens[::-1]
                context = [t[0] for t in tokens]
                prev, prev2 = self.features.START

                max_i = len(tokens) - 1
                for i, (wform, pos) in enumerate(tokens):
                    i_ = max_i - i if rev else i
                    guess, coef = cdict.predict_tag(wform, isfirst=i_ == 0)
                    if self._guess_pos:
                        guess, coef = self._guess_pos(guess, coef, i_,
                                                      tokens_, cdict)
                    if guess is not None:
                        if guess == pos:
                            td2 += 1
                        else:
                            fd2 += 1
                    if guess is None or coef < 1.:
                        features = self.features.get_pos_features(
                            i, context, prev, prev2
                        )
                        guess = model.predict(
                            features, suggest=guess, suggest_coef=coef,
                            dropout=dropout
                        )
                        if guess == pos:
                            tp += 1
                        else:
                            fp += 1
                        model.update(pos, guess, features)
                    elif guess == pos:
                        td += 1
                    else:
                        fd += 1
                    n += 1
                    c += guess == pos

                    prev2 = prev
                    tokens_[i][1] = prev = \
                        guess if not context_dropout \
                              or rand() >= context_dropout else \
                        tags[randint(0, last_tag_idx)]

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    self.evaluate_pos, {'rev': rev}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals, self.evaluate_pos,
            {'rev': rev}
        )

    def train_lemma(self, epochs=5, no_train_evals=True, seed=None):
        """Train a lemma tagger from ``self._train_corpus``.

        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        """
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)

        self._lemma_model = True
        # Yes, we do nothing :)

    def train_feats(self, joint=False, rev=False, feat=None, epochs=5,
                    no_train_evals=True, seed=None, dropout=None,
                    context_dropout=None):
        """Train FEATS taggers from ``self._train_corpus``.

        :param joint: if True, train joint FEATS model; elsewise, train
                      separate models (default)
        :param rev: if True, train Reversed FEATS tagger instead of generic
                    straight one
        :param feat: name of the feat to evaluate the tagger; if None
                     (default), then tagger will be evaluated for all feats
        :type feat: str
        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of FEATS tags to be randomly replaced
                                after predict to random FEATS tags to prevent
                                overfitting
        :type context_dropout: float
        """
        return (
            self._train_feats_joint if joint else
            self._train_feats_separate
        )(
            rev=rev, feat=feat, epochs=epochs, no_train_evals=no_train_evals,
            seed=seed, dropout=dropout, context_dropout=context_dropout
        )

    def _train_feats_separate(self, rev=False, feat=None, epochs=5,
                              no_train_evals=True, seed=None, dropout=None,
                              context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)

        if feat:
            models = self._feats_rev_models if rev else self._feats_models
        else:
            models = {}
            if rev:
                self._feats_rev_models = models
            else:
                self._feats_models = models

        default_val = '_'
        feat_vals = cdict.get_feats()
        if feat:
            feat_vals = {feat: feat_vals[feat]}
        for feat in sorted(feat_vals):
            header = '{}FEAT<<{}>>'.format('Reversed ' if rev else '', feat)
            model = models[feat] = \
                _AveragedPerceptron(default_class=default_val)
            vals = sorted(feat_vals[feat])
            last_val_idx = len(vals) - 1
            print([x for x in vals if x != default_val], file=LOG_FILE)
            best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
            epoch = 0
            while True:
                n = c = 0
                td = fd = td2 = fd2 = tp = fp = 0
                random.shuffle(self._train_corpus)
                print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
                for sent_no, sentence in enumerate(self._train_corpus):
                    if not sent_no % progress_check_step:
                        print_progress(sent_no, end_value=corpus_len,
                                       step=progress_step)

                    #sentence = self.predict_pos(sentence, rev=rev, inplace=False)
                    tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                                  for x in sentence
                                      if x['FORM'] and x['LEMMA'] and x['UPOS']
                                     and '-' not in x['ID']]
                    tokens_ = [[*t[:3], None] for t in tokens]
                    if rev:
                        tokens = tokens[::-1]
                    context, lemma_context, pos_context = \
                        [list(x) for x in zip(*[t[:3] for t in tokens])] \
                            if tokens else \
                        [[]] * 3
                    prev, prev2 = self.features.START

                    max_i = len(tokens) - 1
                    for i, (wform, lemma, pos, feats) in enumerate(tokens):
                        i_ = max_i - i if rev else i
                        gold_val = feats.get(feat, default_val)
                        guess, coef = cdict.predict_feat(feat,
                                                         wform, lemma, pos)
                        if self._guess_feat:
                            guess, coef = self._guess_feat(guess, coef, i_,
                                                           feat, tokens_,
                                                           cdict)
                        if coef is not None:
                            if guess is None:
                                guess = default_val
                            if guess == gold_val:
                                td2 += 1
                            else:
                                fd2 += 1
                        if coef == 1.:
                            if guess == gold_val:
                                td += 1
                            else:
                                fd += 1
                        else:
                            features = self.features.get_feat_features(
                                i, context, lemma_context, pos_context,
                                False, last_val_idx, prev, prev2
                            )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef,
                                dropout=dropout
                            )
                            if guess == gold_val:
                                tp += 1
                            else:
                                fp += 1
                            model.update(gold_val, guess, features)
                        if guess != default_val or gold_val != default_val:
                            n += 1
                            c += guess == gold_val

                        prev2 = prev
                        tokens_[i_][3] = prev = \
                            guess if not context_dropout \
                                  or rand() >= context_dropout else \
                            vals[randint(0, last_val_idx)]

                print_progress(sent_no + 1, end_value=corpus_len,
                               step=progress_step)
                epoch, epochs, best_epoch, best_score, best_weights, \
                                                              eqs, bads, score = \
                    self._train_eval(
                        model, epoch, epochs, epochs_,
                        best_epoch, best_score, best_weights,
                        eqs, bads, score,
                        td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                        lambda **kwargs: self.evaluate_feats(**kwargs)[1],
                        {'joint': False, 'rev': rev, 'feat': feat}
                    )
                if eqs == -1:
                    break

            res = self._train_done(
                header, model, eqs, no_train_evals,
                lambda **kwargs: self.evaluate_feats(**kwargs)[1],
                {'joint': False, 'rev': rev, 'feat': feat}
            )

        return res if feat else \
               f_evaluate(joint=False, rev=rev, feat=feat, silent=True)

    def _train_feats_joint(self, rev=False, feat=None, epochs=5,
                           no_train_evals=True, seed=None, dropout=None,
                           context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)
        assert not feat, 'ERROR: feat must be None with joint=True'

        model = _AveragedPerceptron()
        if rev:
            self._feats_rev_model = model
        else:
            self._feats_model = model

        header = '{}FEATS'.format('Reversed ' if rev else '')
        vals, vals_len, vals_sorted = set(), 0, []
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                              for x in (reversed(sentence) if rev else
                                        sentence)
                                  if x['FORM'] and x['LEMMA'] and x['UPOS']
                                 and '-' not in x['ID']]
                context, lemma_context, pos_context = \
                    [list(x) for x in zip(*[t[:3] for t in tokens])] \
                        if tokens else \
                    [[]] * 3
                prev, prev2 = self.features.START

                max_i = len(tokens) - 1
                for i, (_, _, _, feats) in enumerate(tokens):
                    gold = '|'.join('='.join((x, feats[x]))
                                        for x in sorted(feats))
                    features = self.features.get_feat_features(
                        i, context, lemma_context, pos_context,
                        True, 0, prev, prev2
                    )
                    guess = model.predict(features, dropout=dropout)
                    model.update(gold, guess, features)
                    n += 1
                    c += guess == gold

                    vals.add(gold)
                    if len(vals) > vals_len:
                        vals_sorted = sorted(vals)
                        vals_len = len(vals)
                    prev2 = prev
                    prev = \
                        guess if not context_dropout \
                              or rand() >= context_dropout else \
                        vals_sorted[randint(0, vals_len - 1)]

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    lambda **kwargs: self.evaluate_feats(**kwargs)[1],
                    {'joint': True, 'rev': rev}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals,
            lambda **kwargs: self.evaluate_feats(**kwargs)[1],
            {'joint': True, 'rev': rev}
        )



# -*- coding: utf-8 -*-
# Morra project
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Morra is a part of the RuMor project. It provides tools for complete
morphological sentence parsing and Named-entity recognition.
"""
from morra._version import __version__
from morra.base_parser import autotrain
from morra.morph_parser import MorphParser
from morra.morph_parser2 import MorphParser2
from morra.morph_parser3 import MorphParser3
from morra.morph_parser_ne import MorphParserNE



# -*- coding: utf-8 -*-
# Morra project: Features for MorphParser
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
If you need MorphParser to support your language, add feature-functions for
your language here. Then, create parser as:

    ``MorphParser(features='<your lang>')``
"""
from morra.base_features import BaseFeatures


class Features(BaseFeatures):
    """Features for MorphParser"""

    def __init__(self, lang='RU'):
        super().__init__(lang=lang)
        if lang == 'RU':
            self.get_pos_features = self.get_pos_features_RU
            self.get_feat_features = self.get_feat_features_RU

    def get_pos_features_RU(self, i, context, prev, prev2):
        """Map tokens into a feature representation, implemented as a
        {hashable: int} dict. If the features change, a new model must be
        trained"""
        if prev is None:
            prev = ''
        if prev2 is None:
            prev2 = ''

        context = self.START + context + self.END
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        features = self.init_features()

        # It's useful to have a constant feature, which acts sort of like a prior
        self.add_feature(features, '_')  # bias
        self.add_feature(features, 'w', wform_i)
        self.add_feature(features, 's', self.wform_shape(wform))
        len_ = len(wform)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform[-val:])
        for val in range(3, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform[val:])
        #for val in range(1, 8):
        #    self.add_feature(features, 'P' + str(val),
        #                     '' if val > len_ else wform[:-val])

        #for val in range(1, 8):
        #    self.add_feature(features, '>' + str(val), wform_i[val:val + 2])
        #for val in range(1, 8):
        #    self.add_feature(features, '<' + str(val), wform_i[-val - 2:-val])
        #for val in range(1, 8):
        #    _len = len_ - 1 - val
        #    for start in range(1, _len // 2 + 1):
        #        self.add_feature(features, '>' + str(val) + ':' + str(start),
        #                         wform[start:start + val])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform[len_ -start - val:len_ - start])
        # for val in range(3, 7):
        #     for start in range(1, len_ - 1 - val):
        #         self.add_feature(features, 'i' + str(val) + ':' + str(start),
        #                          wform[start:start + val])
        #for val in range(1, 8):
        #    self.add_feature(features, 'ii' + str(val), wform[val:-val])

        self.add_feature(features, '-1p', prev)
        self.add_feature(features, '-2p', prev2)
        self.add_feature(features, '-1p-2p', prev, prev2)
        self.add_feature(features, '-1pw', prev, wform_i)

        self.add_feature(features, '-1w', wform_b1)
        #self.add_feature(features, '-1s', self.wform_shape(context[i - 1]))
        #self.add_feature(features, '-1s3', wform_b1[-3:])
        self.add_feature(features, '-1s4', wform_b1[-4:])

        self.add_feature(features, '+1w', wform_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])
        #self.add_feature(features, '+1s4', wform_f1[-4:])

        self.add_feature(features, '-2w', wform_b2)

        self.add_feature(features, '+2w', wform_f2)
        return features

    def get_feat_features_RU(self, i, context, lemma_context, pos_context,
                             joint, val_cnt, prev, prev2):
        if prev is None:
            prev = ''
        if prev2 is None:
            prev2 = ''

        context       = self.START +       context + self.END
        lemma_context = self.START + lemma_context + self.END
        pos_context   = self.START +   pos_context + self.END
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        lemma_i  = lemma_context[i]
        lemma_b1 = lemma_context[i - 1]
        lemma_b2 = lemma_context[i - 2]
        lemma_f1 = lemma_context[i + 1]
        lemma_f2 = lemma_context[i + 2]

        pos_i  = pos_context[i]
        pos_b1 = pos_context[i - 1]
        pos_b2 = pos_context[i - 2]
        pos_f1 = pos_context[i + 1]
        pos_f2 = pos_context[i + 2]

        features = self.init_features()

        self.add_feature(features, '_')
        self.add_feature(features, 'w', wform_i)
        self.add_feature(features, 's', self.wform_shape(lemma_i))
        len_ = len(wform)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform[-val:])
        for val in range(3, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform[val:])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform[len_ - start - val:len_ - start])
        self.add_feature(features, 'p', pos_i)

        if joint or val_cnt < 7:
            self.add_feature(features, '-1f', prev)
            self.add_feature(features, '-1fw', prev, wform_i)
            self.add_feature(features, '-1fl', prev, lemma_i)
            self.add_feature(features, '-1fp', prev, pos_i)
        else:
            self.add_feature(features, '-1f-2f', prev, prev2)
            self.add_feature(features, '-1fw', prev, wform_i)
            self.add_feature(features, '-f', prev if prev != '_' else prev2)
            self.add_feature(features, '-fw',
                             prev if prev != '_' else prev2, wform_i)

        self.add_feature(features, '-1l', lemma_b1)
        self.add_feature(features, '-1s4', wform_b1[-4:])
        self.add_feature(features, '-1p', pos_b1)

        self.add_feature(features, '+1l', lemma_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])
        self.add_feature(features, '+1p', pos_f1)

        self.add_feature(features, '-2l', lemma_b2)
        self.add_feature(features, '-2p', pos_b2)

        self.add_feature(features, '+2l', lemma_f2)
        self.add_feature(features, '+2p', pos_f2)
        return features



# -*- coding: utf-8 -*-
# Morra project: Morphological parser 2
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Get the results of forward and backward parsers and make refining parse on the
ground of them.
"""
from collections import OrderedDict
from copy import deepcopy
import pickle
import random
from random import randint, random as rand
import sys

from corpuscula.utils import LOG_FILE, print_progress
from morra.base_parser import _AveragedPerceptron
from morra.features2 import Features2
from morra.morph_parser import MorphParser


class MorphParser2(MorphParser):

    def __init__(self, features='RU',
                 guess_pos=None, guess_lemma=None, guess_feat=None):
        super().__init__(
            guess_pos=guess_pos, guess_lemma=guess_lemma, guess_feat=guess_feat
        )
        self.features = Features2(lang=features) \
            if isinstance(features, str) else features

        self._pos2_model    = None
        self._feats2_model  = None
        self._feats2_models = {}

    def backup(self):
        """Get current state"""
        o = super().backup()
        o.update({'pos2_model_weights'   : self._pos2_model.weights
                                               if self._pos2_model   else
                                           None,
                  'feats2_model_weights' : self._feats2_model.weights
                                               if self._feats2_model else
                                           None,
                  'feats2_models_weights': {
                      x: y.weights for x, y in self._feats2_models.items()
                  }})
        return o

    def restore(self, o):
        """Restore current state from backup object"""
        super().restore(o)
        (pos2_model_weights   ,
         feats2_model_weights ,
         feats2_models_weights) = [o.get(x) for x in ['pos2_model_weights'   ,
                                                      'feats2_model_weights' ,
                                                      'feats2_models_weights']]
        if pos2_model_weights:
            self._pos2_model = _AveragedPerceptron()
            self._pos2_model.weights = pos2_model_weights
        else:
            self._pos2_model = None
        if feats2_model_weights:
            self._feats2_model = _AveragedPerceptron()
            self._feats2_model.weights = feats2_model_weights
        else:
            self._feats2_model = None
        self._feats2_models = {}
        if feats2_models_weights:
            for feat, weights in feats2_models_weights.items():
                model = self._feats2_models[feat] = _AveragedPerceptron()
                model.weights = weights

    def _save_pos2_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._pos2_model.weights if self._pos2_model else
                        None, f, 2)

    def _load_pos2_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._pos2_model = _AveragedPerceptron()
                self._pos2_model.weights = weights
            else:
                self._pos2_model = None

    def _save_feats2_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump(self._feats2_model.weights if self._feats2_model else
                        None, f, 2)

    def _load_feats2_model(self, file_path):
        with open(file_path, 'rb') as f:
            weights = pickle.load(f)
            if weights:
                self._feats2_model = _AveragedPerceptron()
                self._feats2_model.weights = weights
            else:
                self._feats2_model = None

    def _save_feats2_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump(
                (feat, self._feats2_models[feat].weights) if feat else
                {x: y.weights for x, y in self._feats2_models.items()},
                f, 2
            )

    def _load_feats2_models(self, file_path):
        with open(file_path, 'rb') as f:
            o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._feats2_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._feats2_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def predict_pos2(self, sentence, with_backoff=True, max_repeats=0,
                     inplace=True):
        """Tag the *sentence* with the POS-2 tagger.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where POS-1 taggers
                            don't concur
        :type max_repeats: int
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        cdict = self._cdict
        model = self._pos2_model
        assert model, 'ERROR: Use train_pos2() prior to prepare POS-2 tagger'
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        predict = self.predict_pos_ if hasattr(self, 'predict_pos_') else \
                      self.predict_pos
        sent = predict(sent, rev=False, inplace=True)
        sent_rev = predict(sent, rev=True, inplace=False)
        tokens_straight = [(x['FORM'], x['UPOS'])
                               for x in sent
                                   if x['FORM'] and x['UPOS']
                                                and '-' not in x['ID']]
        tokens_rev = [(x['FORM'], x['UPOS'])
                          for x in sent_rev
                              if x['FORM'] and x['UPOS']
                                           and '-' not in x['ID']]
        context, pos_context_straight = \
            [list(x) for x in zip(*[t for t in tokens_straight])] \
                if tokens_straight else \
            [[]] * 2
        pos_context_rev = [t[1] for t in tokens_rev]
        ## Rev model is better for initial word (with capital letter?)
        tokens_ = [[t[0], None] for t in tokens_rev][:2] \
                + [[t[0], None] for t in tokens_straight][2:]
        pos_context = pos_context_rev[:2] + pos_context_straight[2:]
        ###
        changes = len(sent) + 1
        i_ = 1
        while True:
            changes_prev = changes
            changes = 0
            pos_context_straight_i = iter(pos_context_straight)
            pos_context_rev_i      = iter(pos_context_rev     )
            i = 0
            for token in sent:
                wform = token['FORM']
                if wform and '-' not in token['ID']:
                    pos_straight = next(pos_context_straight_i)
                    pos_rev      = next(pos_context_rev_i     )
                    if pos_straight != pos_rev or (not with_backoff
                                               and max_repeats > 0):
                        guess, coef = cdict.predict_tag(wform, isfirst=i == 0)
                        if self._guess_pos:
                            guess, coef = self._guess_pos(guess, coef, i,
                                                          tokens_, cdict)
                        if guess is None or coef < 1.:
                            features = self.features.get_pos2_features(
                                i, context, pos_context
                            )
                            guess = model.predict(
                                features#, suggest=guess, suggest_coef=coef
                            )
                            if with_backoff and guess not in [pos_straight,
                                                              pos_rev]:
                                guess = pos_context[i]
                        if guess != token['UPOS']:
                            changes += 1
                        token['UPOS'] = tokens_[i][1] = pos_context[i] = guess
                    i += 1
            if with_backoff or changes == 0:
                break
            elif changes > changes_prev:
                for token, token_prev in zip(sent, sent_prev):
                    token['UPOS'] = token_prev['UPOS']
                break
            if i_ >= max_repeats:
                break
            sent_prev = deepcopy(sent)
            i_ += 1
        return sentence

    def predict_feats2(self, sentence, joint=False, with_backoff=True,
                       max_repeats=0, feat=None, inplace=True):
        """Tag the *sentence* with the FEATS-2 tagger.

        :param sentence: sentence in Parsed CoNLL-U format; UPOS and LEMMA
                         fields must be already filled
        :type sentence: list(dict)
        :param joint: if True, use joint FEATS-2 model; elsewise, use separate
                      models (default)
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where FEATS-1 taggers
                            don't concur
        :type max_repeats: int
        :param feat: name of the feat to tag; if None, then all possible feats
                     will be tagged
        :type feat: str
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        return (
            self._predict_feats2_joint if joint else
            self._predict_feats2_separate
        )(
            sentence, with_backoff=with_backoff, max_repeats=max_repeats,
            feat=feat, inplace=inplace
        )

    def _predict_feats2_separate(self, sentence, with_backoff=True,
                                 max_repeats=0, feat=None, inplace=True):
        cdict = self._cdict
        models = self._feats2_models
        assert models, \
               'ERROR: Use train_feats2(joint=False) prior to prepare ' \
               'FEATS-2 tagger'
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        if not feat:
            for token in sent:
                token['FEATS'] = OrderedDict()
            for feat in cdict.get_feats():
                self._predict_feats2_separate(sent, with_backoff=with_backoff,
                                              max_repeats=max_repeats,
                                              feat=feat, inplace=True)
        else:
            default_val = '_'
            model = models[feat]
            val_cnt = len(cdict.get_feats()[feat]) - 1
            sent = self._predict_feats_separate(
                sent, rev=False, feat=feat, inplace=True
            )
            sent_rev = self._predict_feats_separate(
                sent, rev=True, feat=feat, inplace=False
            )
            tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                          for x in sent
                              if x['FORM'] and x['LEMMA'] and x['UPOS']
                             and '-' not in x['ID']]
            tokens_rev = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                              for x in sent_rev
                                  if x['FORM'] and x['LEMMA'] and x['UPOS']
                                 and '-' not in x['ID']]
            context, lemma_context, pos_context, feats_context_straight = \
                [list(x) for x in zip(*[t for t in tokens])] if tokens else \
                [[]] * 4
            feats_context_rev = [t[3] for t in tokens_rev]
            # Get straight version as backoff
            tokens_ = [[*t[:3], None] for t in tokens]
            feats_context = deepcopy(feats_context_straight)
            ###
            changes = len(tokens) + 1
            i_ = 1
            while True:
                changes_prev = changes
                changes = 0
                feats_context_straight_i = iter(feats_context_straight)
                feats_context_rev_i      = iter(feats_context_rev     )
                for i, (wform, lemma, pos, feats) in enumerate(tokens):
                    feat_val_straight = next(feats_context_straight_i) \
                                             .get(feat, default_val)
                    feat_val_rev      = next(feats_context_rev_i     ) \
                                             .get(feat, default_val)
                    if feat_val_straight != feat_val_rev or (not with_backoff
                                                         and max_repeats > 0):
                        guess, coef = \
                            cdict.predict_feat(feat, wform, lemma, pos)
                        if self._guess_feat:
                            guess, coef = \
                                self._guess_feat(guess, coef, i, feat,
                                                 tokens_, cdict)
                        if coef is not None and guess is None:
                            guess = default_val
                        if coef != 1.:
                            features = self.features.get_feat2_features(
                                    i, feat, context,
                                    lemma_context, pos_context, feats_context,
                                    False, val_cnt
                                )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef
                            )
                            if with_backoff and guess not in [feat_val_rev,
                                                              feat_val_straight]:
                                guess = feats_context[i].get(feat, default_val)
                        if guess != feats.get(feat, default_val):
                            changes += 1
                        tokens_[i][3] = guess
                        if guess != default_val:
                            feats[feat] = feats_context[i][feat] = guess
                        else:
                            feats.pop(feat, None)
                            feats_context[i].pop(feat, None)
                if with_backoff or changes == 0:
                    break
                elif changes > changes_prev:
                    for token, token_prev in zip(tokens, tokens_prev):
                        tokens[3] = token_prev[3].copy()
                    break
                if i_ >= max_repeats:
                    break
                tokens_prev = deepcopy(tokens)
                i_ += 1
        return sentence

    def _predict_feats2_joint(self, sentence, with_backoff=True, feat=None,
                              max_repeats=0, inplace=True):
        assert not feat, 'ERROR: feat must be None with joint=True'
        cdict = self._cdict
        model = self._feats2_model
        assert model, \
               'ERROR: Use train_feats2(joint=True) prior to prepare ' \
               'FEATS-2 tagger'
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        sent = self._predict_feats_joint(sent, rev=False, inplace=True)
        sent_rev = self._predict_feats_joint(sent, rev=True, inplace=False)
        tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                      for x in sent
                          if x['FORM'] and x['LEMMA'] and x['UPOS']
                         and '-' not in x['ID']]
        tokens_rev = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                          for x in sent_rev
                              if x['FORM'] and x['LEMMA'] and x['UPOS']
                             and '-' not in x['ID']]
        context, lemma_context, pos_context, feats_context_straight = \
            [list(x) for x in zip(*[t for t in tokens])] if tokens else \
            [[]] * 4
        feats_context_rev = [t[3] for t in tokens_rev]
        # Rev model is better for initial word (with capital letter?)
        feats_context = deepcopy(feats_context_rev[:1]
                               + feats_context_straight[1:])
        ###
        changes = len(feats_context_straight) + 1
        i_ = 1
        while True:
            changes_prev = changes
            changes = 0
            feats_context_rev_i      = iter(feats_context_rev     )
            feats_context_straight_i = iter(feats_context_straight)
            for i, feats in enumerate(feats_context_straight):
                feats_rev = next(feats_context_rev_i)
                feat_vals_rev = \
                    '|'.join('='.join((x, feats_rev[x]))
                                 for x in sorted(feats_rev))
                feats_straight = next(feats_context_straight_i)
                feat_vals_straight = \
                    '|'.join('='.join((x, feats_straight[x]))
                                 for x in sorted(feats_straight))
                if feat_vals_rev != feat_vals_straight or (not with_backoff
                                                       and max_repeats > 0):
                    features = self.features.get_feat2_features(
                        i, None,
                        context, lemma_context, pos_context, feats_context,
                        True, 0
                    )
                    guess = model.predict(features)
                    feats_ctx = '|'.join('='.join((x, feats_context[i][x]))
                                             for x in sorted(feats_context[i]))
                    if with_backoff and guess not in [feat_vals_rev,
                                                      feat_vals_straight]:
                        guess = feats_ctx
                    elif guess != feats_ctx:
                        changes += 1
                    feats.clear()
                    feats_ctx = feats_context[i]
                    feats_ctx.clear()
                    if guess:
                        for feat, val in [t.split('=') for t in guess.split('|')]:
                            feats[feat] = feats_ctx[feat] = val
            if with_backoff or changes == 0:
                break
            elif changes > changes_prev:
                for feats, feats_prev in zip(feats_context_straight,
                                             feats_context_straight_prev):
                    feats.clear()
                    feats.update(feats_prev)
                break
            if i_ >= max_repeats:
                break
            feats_context_straight_prev = deepcopy(feats_context_straight)
            i_ += 1
        return sentence

    def predict2(self, sentence, pos_backoff=True, pos_repeats=0,
                 feats_joint=False, feats_backoff=True, feats_repeats=0,
                 inplace=True):
        """Tag the *sentence* with the all available taggers.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one
                            repeat - only for tokens where POS-1 taggers
                            don't concur
        :type pos_repeats: int
        :param feats_joint: if True, use joint model; elsewise, use separate
                            models (default)
        :type feats_backoff: if result of FEATS-2 tagger differs from both its
                             base taggers, get one of the bases on the ground
                             of some heuristics
        :param feats_repeats: repeat a prediction step based on the previous
                              one while changes in prediction are diminishing
                              and ``max_repeats`` is not reached. 0 (default)
                              means one repeat - only for tokens where FEATS-1
                              taggers don't concur
        :type feats_repeats: int
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        return \
            self.predict_feats2(
                self.predict_lemma(
                    self.predict_pos2(
                        sentence, with_backoff=pos_backoff,
                        max_repeats=pos_repeats, inplace=inplace
                    ),
                    inplace=inplace
                ),
                joint=feats_joint, with_backoff=feats_backoff,
                max_repeats=feats_repeats, inplace=inplace
            )

    def predict_pos2_sents(self, sentences=None, with_backoff=True,
                           max_repeats=0, inplace=True, save_to=None):
        """Apply ``self.predict_pos2()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where POS-1 taggers
                            don't concur
        :type max_repeats: int
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_pos2(
                     s, with_backoff=with_backoff, max_repeats=max_repeats,
                     inplace=inplace
                 )
                     for s in sentences),
            save_to=save_to
        )

    def predict_feats2_sents(self, sentences=None, joint=False,
                             with_backoff=True, max_repeats=0, feat=None,
                             inplace=True, save_to=None):
        """Apply ``self.predict_feats2()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param joint: if True, use joint FEATS-2 model; elsewise, use separate
                      models (default)
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where FEATS-1 taggers
                            don't concur
        :type max_repeats: int
        :param feat: name of the feat to tag; if None, then all feats will be
                     tagged
        :type feat: str
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, the new list of
                        sentences will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_feats2(
                     s, joint=joint, with_backoff=with_backoff,
                     max_repeats=max_repeats, feat=feat, inplace=inplace
                 )
                     for s in sentences),
            save_to=save_to
        )

    def predict2_sents(self, sentences=None, pos_backoff=True, pos_repeats=0,
                       feats_joint=False, feats_backoff=True, feats_repeats=0,
                       inplace=True, save_to=None):
        """Apply ``self.predict2()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one
                            repeat - only for tokens where POS-1 taggers
                            don't concur
        :type pos_repeats: int
        :param feats_joint: if True, use joint model; elsewise, use separate
                            models (default)
        :type feats_backoff: if result of FEATS-2 tagger differs from both its
                            base taggers, get one of the bases on the ground
                            of some heuristics
        :param feats_repeats: repeat a prediction step based on the previous
                              one while changes in prediction are diminishing
                              and ``max_repeats`` is not reached. 0 (default)
                              means one repeat - only for tokens where FEATS-1
                              taggers don't concur
        :type feats_repeats: int
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict2(
                     s, pos_backoff=pos_backoff, pos_repeats=pos_repeats,
                     feats_joint=feats_joint, feats_backoff=feats_backoff,
                     feats_repeats=feats_repeats, inplace=inplace
                 )
                     for s in sentences),
            save_to=save_to
        )

    def evaluate_pos2(self, gold=None, test=None, with_backoff=True,
                      max_repeats=0, pos=None, unknown_only=False,
                      silent=False):
        """Score the accuracy of the POS tagger against the *gold* standard.
        Remove POS tags from the *gold* standard text, retag it using the
        tagger, then compute the accuracy score. If *test* is not None, compute
        the accuracy of the *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where POS-1 taggers
                            don't concur
        :type max_repeats: int
        :param pos: name of the tag to evaluate the tagger; if None, then
                    tagger will be evaluated for all tags
        :type pos: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy score of the tagger against the gold
        :rtype: float
        """
        self.predict_pos_ = self.predict_pos
        self.predict_pos = \
            lambda sentence, rev=None, inplace=True: \
                self.predict_pos2(sentence, with_backoff=with_backoff,
                                  max_repeats=max_repeats, inplace=inplace)
        res = self.evaluate_pos(gold=gold, test=test, pos=pos,
                                unknown_only=unknown_only, silent=silent)
        self.predict_pos = self.predict_pos_
        del self.predict_pos_
        return res

    def evaluate_feats2(self, gold=None, test=None, joint=False,
                        with_backoff=True, max_repeats=0,
                        feat=None, unknown_only=False, silent=False):
        """Score the accuracy of the FEATS-2 tagger against the *gold*
        standard. Remove feats (or only one specified feat) from the *gold*
        standard text, generate new feats using the tagger, then compute the
        accuracy score. If *test* is not None, compute the accuracy of the
        *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param joint: if True, use joint FEATS-2 model; elsewise, use separate
                      models (default)
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 (default) means
                            one repeat - only for tokens where FEATS-1 taggers
                            don't concur
        :type max_repeats: int
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy scores of the tagger against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not;
                 2. by tags: sum of correctly detected feats to sum of all
                    feats that are non-empty in either gold or retagged 
                    sentence
        :rtype: tuple(float, float)
        """
        f = self.predict_feats
        self.predict_feats = \
            lambda sentence, joint=joint, rev=None, feat=feat, inplace=True: \
                self.predict_feats2(sentence, joint=joint,
                                    with_backoff=with_backoff,
                                    max_repeats=max_repeats, feat=feat,
                                    inplace=inplace)
        res = self.evaluate_feats(gold=gold, test=test, joint=joint, feat=feat,
                                  unknown_only=unknown_only, silent=silent)
        self.predict_feats = f
        return res

    def evaluate2(self, gold=None, test=None, pos_backoff=True, pos_repeats=0,
                  feats_joint=False, feats_backoff=True, feats_repeats=0,
                  feat=None, unknown_only=False, silent=False):
        """Score a joint accuracy of the all available taggers against the
        *gold* standard. Extract wforms from the *gold* standard text, retag it
        using all the taggers, then compute a joint accuracy score. If *test*
        is not None, compute the accuracy of the *test* corpus with respect to
        the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one
                            repeat - only for tokens where POS-1 taggers
                            don't concur
        :type pos_repeats: int
        :param feats_joint: if True, use joint model; elsewise, use separate
                            models (default)
        :type feats_backoff: if result of FEATS-2 tagger differs from both its
                            base taggers, get one of the bases on the ground
                            of some heuristics
        :param feats_repeats: repeat a prediction step based on the previous
                              one while changes in prediction are diminishing
                              and ``max_repeats`` is not reached. 0 (default)
                              means one repeat - only for tokens where FEATS-1
                              taggers don't concur
        :type feats_repeats: int
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: joint accuracy scores of the taggers against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not
                 2. by tags: sum of correctly detected tags to sum of all tags
                    that are non-empty in either gold or retagged sentences
        :rtype: tuple(float, float)
        """
        f = self.predict
        self.predict = \
            lambda sentence, pos_rev=None, \
                   feats_joint=feats_joint, feats_rev=None, inplace=False: \
                self.predict2(
                    sentence, pos_backoff=pos_backoff, pos_repeats=pos_repeats,
                    feats_joint=feats_joint, feats_backoff=feats_backoff,
                    feats_repeats=feats_repeats, inplace=inplace
                )
        res = self.evaluate(gold=gold, test=test, feats_joint=feats_joint,
                            feat=feat, unknown_only=unknown_only,
                            silent=silent)
        self.predict = f
        return res

    def train_pos2(self, epochs=5, test_max_repeats=0, no_train_evals=True,
                   seed=None, dropout=None, context_dropout=None):
        """Train a POS-2 tagger from ``self._train_corpus``.

        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param test_max_repeats: parameter for ``evaluate_pos2()``
        :type test_max_repeats: int
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of POS tags to be randomly replaced
                                after predict to random POS tags to prevent
                                overfitting
        :type context_dropout: float
        """
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)

        assert self._pos_model, \
               'ERROR: Use train_pos() prior to prepare POS tagger'
        assert self._pos_rev_model, \
               'ERROR: Use train_pos(rev=True) prior to prepare ' \
               'Reversed POS tagger'

        model = self._pos2_model = \
            _AveragedPerceptron(default_class=cdict.most_common_tag())

        header = 'POS-2'
        tags = sorted(cdict.get_tags())
        last_tag_idx = len(tags) - 1
        print(tags, file=LOG_FILE)
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['UPOS'])
                              for x in sentence
                                  if x['FORM'] and '-' not in x['ID']]
                context, pos_context = \
                    [list(x) for x in zip(*[t for t in tokens])] \
                        if tokens else \
                    [[]] * 2
                tokens_ = [[t[0], None] for t in tokens]

                for i, (wform, pos) in enumerate(tokens):
                    guess, coef = cdict.predict_tag(wform, isfirst=i == 0)
                    if self._guess_pos:
                        guess, coef = self._guess_pos(guess, coef, i,
                                                      tokens_, cdict)
                    if guess is not None:
                        if guess == pos:
                            td2 += 1
                        else:
                            fd2 += 1
                    if guess is None or coef < 1.:
                        features = self.features.get_pos2_features(
                            i, context, pos_context
                        )
                        guess = model.predict(
                            features,# suggest=guess, suggest_coef=coef,
                            dropout=dropout
                        )
                        if guess == pos:
                            tp += 1
                        else:
                            fp += 1
                        model.update(pos, guess, features)
                    elif guess == pos:
                        td += 1
                    else:
                        fd += 1
                    n += 1
                    c += guess == pos

                    tokens_[i][1] = pos_context[i] = \
                        guess if not context_dropout \
                              or rand() >= context_dropout else \
                        tags[randint(0, last_tag_idx)]

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    self.evaluate_pos2,
                    {'with_backoff': False, 'max_repeats': test_max_repeats}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals, self.evaluate_pos2,
            {'with_backoff': False, 'max_repeats': test_max_repeats}
        )

    def train_feats2(self, joint=False, feat=None, epochs=5,
                     test_max_repeats=0, no_train_evals=True, seed=None,
                     dropout=None, context_dropout=None):
        """Train FEATS-2 taggers from ``self._train_corpus``.

        :param joint: if True, use joint FEATS-2 model; elsewise, train
                      separate models (default)
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param test_max_repeats: parameter for ``evaluate_feats2()``
        :type test_max_repeats: int
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of FEATS tags to be randomly replaced
                                after predict to random FEATS tags to prevent
                                overfitting
        :type context_dropout: float
        """
        return (
            self._train_feats2_joint if joint else
            self._train_feats2_separate
        )(
            feat=feat, epochs=epochs,
            no_train_evals=no_train_evals, test_max_repeats=test_max_repeats,
            seed=seed, dropout=dropout, context_dropout=context_dropout
        )

    def _train_feats2_separate(self, feat=None, epochs=5,
                               no_train_evals=True, test_max_repeats=0,
                               seed=None, dropout=None, context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)

        assert self._feats_models, \
               'ERROR: Use train_feats() prior to prepare FEATS tagger'
        assert self._feats_rev_models, \
               'ERROR: Use train_feats(rev=True) prior to prepare ' \
               'Reversed FEATS tagger'

        if feat:
            models = self._feats2_models
        else:
            models = self._feats2_models = {}

        default_val = '_'
        feat_vals = cdict.get_feats()
        if feat:
            feat_vals = {feat: feat_vals[feat]}
        for feat in sorted(feat_vals):
            header = 'FEAT-2<<{}>>'.format(feat)
            model = models[feat] = \
                _AveragedPerceptron(default_class=default_val)
            vals = sorted(feat_vals[feat])
            last_val_idx = len(vals) - 1
            print([x for x in vals if x != default_val], file=LOG_FILE)
            best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
            epoch = 0
            while True:
                n = c = 0
                td = fd = td2 = fd2 = tp = fp = 0
                random.shuffle(self._train_corpus)
                print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
                for sent_no, sentence in enumerate(self._train_corpus):
                    if not sent_no % progress_check_step:
                        print_progress(sent_no, end_value=corpus_len,
                                       step=progress_step)

                    tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                                  for x in sentence
                                      if x['FORM'] and x['LEMMA'] and x['UPOS']
                                     and '-' not in x['ID']]
                    context, lemma_context, pos_context, feats_context = \
                        [list(x) for x in zip(*[t for t in tokens])] \
                            if tokens else \
                        [[]] * 4
                    tokens_ = [[*t[:3], None] for t in tokens]

                    for i, (wform, lemma, pos, feats) in enumerate(tokens):
                        gold_val = feats.get(feat, default_val)
                        guess, coef = \
                            self._cdict.predict_feat(feat,
                                                     wform, lemma, pos)
                        if self._guess_feat:
                            guess, coef = self._guess_feat(guess, coef, i,
                                                           feat, tokens_,
                                                           cdict)
                        if coef is not None:
                            if guess is None:
                                guess = default_val
                            if guess == gold_val:
                                td2 += 1
                            else:
                                fd2 += 1
                        if coef == 1.:
                            if guess == gold_val:
                                td += 1
                            else:
                                fd += 1
                        else:
                            features = self.features.get_feat2_features(
                                i, feat, context,
                                lemma_context, pos_context, feats_context,
                                False, last_val_idx
                            )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef,
                                dropout=dropout
                            )
                            if guess == gold_val:
                                tp += 1
                            else:
                                fp += 1
                            model.update(gold_val, guess, features)
                        if guess != default_val or gold_val != default_val:
                            n += 1
                            c += guess == gold_val

                        tokens_[i][3] = \
                            guess if not context_dropout \
                                  or rand() >= context_dropout else \
                            vals[randint(0, last_val_idx)]

                print_progress(sent_no + 1, end_value=corpus_len,
                               step=progress_step)
                epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                    self._train_eval(
                        model, epoch, epochs, epochs_,
                        best_epoch, best_score, best_weights,
                        eqs, bads, score,
                        td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                        lambda **kwargs: self.evaluate_feats2(**kwargs)[1],
                        {'joint': False, 'with_backoff': False,
                         'max_repeats': test_max_repeats, 'feat': feat}
                    )
                if eqs == -1:
                    break

            res = self._train_done(
                header, model, eqs, no_train_evals,
                lambda **kwargs: self.evaluate_feats2(**kwargs)[1],
                {'joint': False, 'with_backoff': False,
                 'max_repeats': test_max_repeats, 'feat': feat}
            )

        return res if feat else \
               f_evaluate(joint=False, rev=rev, feat=feat, silent=True)

    def _train_feats2_joint(self, feat=None, epochs=5,
                            no_train_evals=True, test_max_repeats=0,
                            seed=None, dropout=None, context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                              epochs, epochs_ = self._train_init(epochs, seed)
        assert not feat, 'ERROR: feat must be None with joint=True'
        assert not context_dropout, \
               'ERROR: context_dropout must be None with joint=True'

        assert self._feats_model, \
               'ERROR: Use train_feats(joint=True) prior to prepare ' \
               'joint FEATS tagger'
        assert self._feats_rev_model, \
               'ERROR: Use train_feats(joint=True, rev=True) prior ' \
               'to prepare Reversed joint FEATS tagger'

        model = self._feats2_model = _AveragedPerceptron(default_class='')

        header = 'FEATS-2'
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'])
                              for x in sentence
                                  if x['FORM'] and x['LEMMA'] and x['UPOS']
                                 and '-' not in x['ID']]
                context, lemma_context, pos_context, feats_context = \
                    [list(x) for x in zip(*[t for t in tokens])] \
                        if tokens else \
                    [[]] * 4

                for i, feats in enumerate(feats_context):
                    gold = '|'.join('='.join((x, feats[x]))
                                        for x in sorted(feats))
                    features = self.features.get_feat2_features(
                        i, None,
                        context, lemma_context, pos_context, feats_context,
                        True, 0
                    )
                    guess = model.predict(features, dropout=dropout)
                    model.update(gold, guess, features)
                    n += 1
                    c += guess == gold

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    lambda **kwargs: self.evaluate_feats2(**kwargs)[1],
                    {'joint': True, 'with_backoff': False,
                     'max_repeats': test_max_repeats}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals,
            lambda **kwargs: self.evaluate_feats2(**kwargs)[1],
            {'joint': True, 'with_backoff': False,
            'max_repeats': test_max_repeats}
        )



# -*- coding: utf-8 -*-
# Morra project: Base features
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
from collections import OrderedDict

class BaseFeatures:
    """Base class for all features of the project"""

    START = ['-START2-', '-START-']
    END = ['-END-', '-END2-']
    alphabet_RU = ''.join(['ё']
                        + [chr(i) for i in range(ord('а'), ord('я') + 1)])

    def __init__(self, lang='RU'):
        if lang == 'RU':
            self.alphabet = self.alphabet_RU
            self.alphabet_upper = self.alphabet.upper()
        else:
            raise ValueError('ERROR: Features for lang "{}" are not defined'
                                 .format(lang))

    @staticmethod
    def init_features():
        return OrderedDict()

    @staticmethod
    def add_feature(features, name, *args):
        f = ' '.join((name,) + tuple(args))
        features[f] = features.get(f, 0) + 1

    def normalize(self, wform):
        '''
        Normalization used in pre-processing.
        - All words are lower cased
        - Groups of digits of length 4 are represented as !YEAR;
        - Other digits are represented as !DIGITS

        :rtype: str
        '''
        #if '-' in wform and wform[0] != '-':
        #    return '!HYPHEN'
        #elif word.isdigit() and len(wform) == 4:
        #    return '!YEAR'
        #el
        if wform is None:
            return wform
        elif wform[0].isdecimal():
            return '!DIGITS'
        else:
            return wform.lower()

    def wform_shape(self, wform):
        shape = ''
        for c in wform:
            if c in self.alphabet:
                c = 'x'
            elif c in self.alphabet_upper:
                c = 'X'
            elif c.isalpha():
                if c.isupper():
                    c = 'A'
                else:
                    c = 'a'
            elif c.isnumeric():
                c = 'd'
            shape += c
        if len(shape) > 5:
            shape = shape[:2] + ''.join(sorted(set(shape[2:-2]))) + shape[-2:]
        return shape



# -*- coding: utf-8 -*-
# Morra project: Features for MorphParserNE
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
If you need MorphParserNE to support your language, add feature-functions for
your language here. Then, create parser as:

    ``MorphParserNE(features='<your lang>')``
"""
from morra.base_features import BaseFeatures


class FeaturesNE(BaseFeatures):
    """Features for MorphParserNE"""

    def __init__(self, lang='RU'):
        super().__init__(lang=lang)
        if lang == 'RU':
            self.get_ne_features = self.get_ne_features_RU
            self.get_ne2_features = self.get_ne2_features_RU

    def get_ne_features_RU(self, i, context, lemma_context, pos_context,
                           feats_context, prev, prev2):
        prev  = str(prev)
        prev2 = str(prev2)

        context       = self.START +       context + self.END
        lemma_context = self.START + lemma_context + self.END
        pos_context   = self.START +   pos_context + self.END
        feats_context =    [{}, {}] + feats_context + [{}, {}]
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        lemma_i  = lemma_context[i]
        lemma_b1 = lemma_context[i - 1]
        lemma_b2 = lemma_context[i - 2]
        lemma_f1 = lemma_context[i + 1]
        lemma_f2 = lemma_context[i + 2]

        pos_i  = pos_context[i]
        pos_b1 = pos_context[i - 1]
        pos_b2 = pos_context[i - 2]
        pos_f1 = pos_context[i + 1]
        pos_f2 = pos_context[i + 2]

        feats_i  = feats_context[i]
        feats_b1 = feats_context[i - 1]
        feats_b2 = feats_context[i - 2]
        feats_f1 = feats_context[i + 1]
        feats_f2 = feats_context[i + 2]

        features = self.init_features()

        self.add_feature(features, '_')
        self.add_feature(features, 'w', wform_i)
        self.add_feature(features, 's', self.wform_shape(lemma_i))
        len_ = len(wform)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform[-val:])
        for val in range(3, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform[val:])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform[len_ - start - val:len_ - start])
        self.add_feature(features, 'p', pos_i)
        #for feat, val in feats_i.items():
        #    self.add_feature(features, 'i feat', feat, val)
        #case = feats_i.get('Case')
        #if case:
        #    self.add_feature(features, 'i feat-Case', case)

        self.add_feature(features, '-1n', prev)
        self.add_feature(features, '-2n', prev2)
        self.add_feature(features, '-1n-2n', prev, prev2)
        self.add_feature(features, '-1nl', prev, lemma_i)

        self.add_feature(features, '-1l', lemma_b1)
        self.add_feature(features, '-1s4', wform_b1[-4:])
        self.add_feature(features, '-1p', pos_b1)

        self.add_feature(features, '+1l', lemma_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])
        self.add_feature(features, '+1p', pos_f1)

        self.add_feature(features, '-2l', lemma_b2)
        self.add_feature(features, '-2p', pos_b2)

        self.add_feature(features, '+2l', lemma_f2)
        self.add_feature(features, '+2p', pos_f2)
        return features

    def get_ne2_features_RU(self, i, context, lemma_context, pos_context,
                            feats_context, ne_context):
        context       = self.START +       context + self.END
        lemma_context = self.START + lemma_context + self.END
        pos_context   = self.START +   pos_context + self.END
        feats_context =    [{}, {}] + feats_context + [{}, {}]
        ne_context    = self.START +    ne_context + self.END
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        lemma_i  = lemma_context[i]
        lemma_b1 = lemma_context[i - 1]
        lemma_b2 = lemma_context[i - 2]
        lemma_f1 = lemma_context[i + 1]
        lemma_f2 = lemma_context[i + 2]

        pos_i  = pos_context[i]
        pos_b1 = pos_context[i - 1]
        pos_b2 = pos_context[i - 2]
        pos_f1 = pos_context[i + 1]
        pos_f2 = pos_context[i + 2]

        feats_i  = feats_context[i]
        feats_b1 = feats_context[i - 1]
        feats_b2 = feats_context[i - 2]
        feats_f1 = feats_context[i + 1]
        feats_f2 = feats_context[i + 2]

        ne_i  = str(ne_context[i])
        ne_b1 = str(ne_context[i - 1])
        ne_b2 = str(ne_context[i - 2])
        ne_f1 = str(ne_context[i + 1])
        ne_f2 = str(ne_context[i + 2])

        features = self.init_features()

        self.add_feature(features, '_')
        self.add_feature(features, 'w', wform)
        self.add_feature(features, 's', self.wform_shape(wform))
        len_ = len(wform_i)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform_i[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform_i[-val:])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform_i[len_ - start - val:len_ - start]) 
        for val in range(2, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform_i[val:])
        self.add_feature(features, 'p', pos_i)

        self.add_feature(features, '-1n', ne_b1)
        self.add_feature(features, '-2n', ne_b2)
        self.add_feature(features, '-1n-2n', ne_b1, ne_b2)
        self.add_feature(features, '-1nl', ne_b1, lemma_i)
        self.add_feature(features, '+1n', ne_f1)
        self.add_feature(features, '+2n', ne_f2)
        self.add_feature(features, '+1n+2n', ne_f1, ne_f2)
        self.add_feature(features, '+1nl', ne_f1, lemma_i)
        self.add_feature(features, '-1n+1n', ne_b1, ne_f1)

        self.add_feature(features, '-1l', lemma_b1)
        self.add_feature(features, '-1s4', wform_b1[-4:])
        self.add_feature(features, '-1p', pos_b1)

        self.add_feature(features, '+1l', lemma_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])
        self.add_feature(features, '+1p', pos_f1)

        self.add_feature(features, '-2l', lemma_b2)
        self.add_feature(features, '-2pos', pos_b2)

        self.add_feature(features, '+2l', lemma_f2)
        self.add_feature(features, '+2p', pos_f2)
        return features



import os
from setuptools import setup, find_packages

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
version_file_contents = open(os.path.join(SCRIPT_DIR, 'morra/_version.py'),
                             'rt', encoding='utf-8').read()
VERSION = version_file_contents.strip()[len('__version__ = "'):-1]

setup(
    name='morra',
    version=VERSION,
    description='Morphological parser (POS, lemmata, NER etc.)',
    long_description=open('README.md', 'rt', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Sergei Ternovykh',
    author_email='fostroll@gmail.com',
    url='https://github.com/fostroll/morra',
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    # What does your project relate to?
    keywords='natural-language-processing nlp morphology',

    packages=find_packages(exclude=['doc', 'examples', 'scripts', 'tests']),
    install_requires=['corpuscula', 'pymorphy2', 'scipy'],
    include_package_data=True,
    python_requires='>=3.5',
)



# -*- coding: utf-8 -*-
# Morra project
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
A tool that allows use several MorphParsers together in the ensemble.
"""
from copy import deepcopy

from corpuscula.corpus_utils import _AbstractCorpus
from corpuscula.utils import LOG_FILE, print_progress
from morra.base_parser import BaseParser

class MorphEnsemble:
    """A simple tool that allow to conjoin several MorphParsers to the
    ensemble"""

    def __init__(self, cdict):
        """
        :param cdict: corpuscula CorpusDict of the current training corpus
        """
        self._cdict = cdict
        self._predict_methods = []

    def add(self, predict_method, **kwargs):
        """Add parser to the ensemble.

        :param predict_method: a method of MorphParser that we prefer to use
                               for prediction
        :param kwargs: keyword arguments for *predict_method*
        :return: index of added method
        """
        index = len(self._predict_methods)
        self._predict_methods.append((predict_method, kwargs))
        return index

    def pop(self, index):
        """Remove parser from the ensemble.

        :param index: an index returned by the `.add()` method
        :return: the method removed
        """
        return self.models.pop(index)

    def predict(self, fields_to_predict, sentence, inplace=True):
        """Predict necessary fields of *sentences*.

        :param fields_to_predict: the list of CoNLL-U fields names you want to
                                  get a prediction for. E.g.: ['UPOS',
                                  'LEMMA', 'FEATS]. Note, that these fields
                                  must be between the fields that the methods
                                  added with the `.add()` can predict
        :param sentence: the sentence in *Parsed CoNLL-U* format
        :param inplace: if True (default), method changes and returns the given
                        *sentence** itself. Elsewise, the new sentence will be
                        created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        if isinstance(fields_to_predict, str):
            fields_to_predict = [fields_to_predict]
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        results = []
        for predict, kwargs in self._predict_methods:
            results.append((x for x in predict(deepcopy(sent), **kwargs)))
        results_len = len(results)

        for token in sent:
            for fld in fields_to_predict:
                res = {}
                if isinstance(token[fld], dict):
                    best_val, best_scores = {}, {}
                    for s in results:
                        t = next(s)
                        feats = t[fld]
                        for feat, val in feats.items():
                            score = res[feat][val] = \
                                res.setdefault(feat, {None: results_len}) \
                                   .get(val, 0) + 1
                            res[feat][None] -= 1
                            if score > best_scores.get(feat, (None, -1))[1]:
                                best_scores[feat] = (val, score)
                    for feat, scores in best_scores.items():
                        val, score = scores
                        if score >= res[feat][None]:
                            best_val[feat] = val
                else:
                    best_val, best_score = None, -1
                    for s in results:
                        t = next(s)
                        tag = t[fld]
                        score = res[tag] = res.setdefault(tag, 0) + 1
                        if score > best_score:
                            best_val, best_score = tag, score
                token[fld] = best_val
        return sentence

    def predict_sents(self, fields_to_predict, sentences, inplace=True,
                      save_to=None):
        """Apply ``self.predict()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        if isinstance(sentences, type) and issubclass(sentences,
                                                      _AbstractCorpus):
            sentences = sentences.test()
        assert sentences, 'ERROR: Sentences must not be empty'
        sentences = BaseParser._get_corpus(sentences, asis=True)
        sentences = self.predict(fields_to_predict, sentences, inplace=inplace)
        if save_to:
            self.save_conllu(sentences, save_to, silent=True)
            sentences = BaseParser._get_corpus(save_to, asis=True)
        return sentences

    def evaluate(self, fields_to_evaluate, gold=None, test=None,
                 feat=None, unknown_only=False, silent=False):
        """Score an accuracy of the ensemble tagging against the *gold*
        standard. Extract wforms from the *gold* standard text, retag it using
        the ensemble, then compute a joint accuracy score. If *test* is not
        None, compute the accuracy of the *test* corpus with respect to the
        gold.

        :param fields_to_evaluate: the list of CoNLL-U fields names you want to
                                   evaluate a prediction for. E.g.: ['UPOS',
                                   'LEMMA', 'FEATS]. Note, that these fields
                                   must be between the fields that the methods
                                   added with the `.add()` can predict
        :param gold: a corpus of tagged sentences to score the tagger on.
        :param test: a corpus of tagged sentences to compare with *gold*
        :param feat: name of the feat to evaluate the ensemble; if None, then
                     the ensemble will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that not
                             present in train corpus
        :param silent: suppress log
        :return: joint accuracy scores of the taggers against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not
                 2. by tags: sum of correctly detected tags to sum of all tags
                    that are non-empty in either gold or retagged sentences
        :rtype: tuple(float, float)
        """
        n = c = nt = ct = 0
        if (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
        or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = BaseParser._get_corpus(gold, silent=True)
        if test:
            test = BaseParser._get_corpus(test, silent=True)
        if not silent:
            print('Evaluate', file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        cdict = self._cdict
        feat_vals = cdict.get_feats()
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            test_sent = next(test) if test else \
                        self.predict(fields_to_evaluate, gold_sent,
                                     inplace=False)
            for j, gold_token in enumerate(gold_sent):
                wform = gold_token['FORM']
                if wform and '-' not in gold_token['ID']:
                    gold_pos = gold_token['UPOS']
                    if not unknown_only or cdict.wform_isknown(wform,
                                                               tag=gold_pos):
                        test_token = test_sent[j]
                        nf = cf = 0
                        # LEMMA
                        if 'LEMMA' in fields_to_evaluate:
                            nf += 1
                            cf += gold_token['LEMMA'] == test_token['LEMMA']
                        # UPOS
                        if 'UPOS' in fields_to_evaluate:
                            nf += 1
                            cf += gold_pos == test_token['UPOS']
                        # FEATS
                        if 'FEATS' in fields_to_evaluate:
                            gold_feats = gold_token['FEATS']
                            test_feats = test_token['FEATS']
                            for feat_ in [feat] if feat else feat_vals:
                                gold_feat = gold_feats.get(feat_)
                                test_feat = test_feats.get(feat_)
                                if gold_feat or test_feat:
                                    nf += 1
                                    if gold_feat == test_feat:
                                        cf += 1
                        if nf > 0:
                            n += 1
                            c += nf == cf
                            nt += nf
                            ct += cf
        res = c / n if n > 0 else 1., ct / nt if nt > 0 else 1.
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                print('   total: {} tokens, {} tags'.format(n, nt),
                      file=LOG_FILE)
                print(' correct: {} tokens, {} tags'.format(c, ct),
                      file=LOG_FILE)
                print('   wrong: {} tokens, {} tags'.format(n - c, nt - ct),
                      file=LOG_FILE)
                print('Accuracy: {} / {}'.format(res[0], res[1]),
                      file=LOG_FILE)
        return res
